/*
 * Copyright (c) 2016 Google, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/**
 * Truth is an open source, fluent testing framework for Java that is designed to make your test
 * assertions and failure messages more readable.
 *
 * <p>This package is a part of the open-source <a href="http://github.com/google/truth">Truth</a>
 * library.
 */
@CheckReturnValue
@ParametersAreNonnullByDefault
package com.google.common.truth;

import javax.annotation.CheckReturnValue;
import javax.annotation.ParametersAreNonnullByDefault;
